﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Drawing;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Hive;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Clients.Hive.Jobs {
  [Item("Optimizer Job", "Represents Job which executes a IOptimizer object.")]
  [StorableClass]
  public class OptimizerJob : AbstractJob {
    public override bool IsParallelizable {
      get { return this.Optimizer is Optimization.Experiment || this.Optimizer is BatchRun; }
    }

    [Storable]
    protected IOptimizer optimizer;
    public IOptimizer Optimizer {
      get { return optimizer; }
      set {
        if (value != optimizer) {
          if (optimizer != null) DeregisterOptimizerEvents();
          optimizer = value;
          if (optimizer != null) RegisterOptimizerEvents();
          OnOptimizerChanged();
        }
      }
    }

    [Storable]
    private int indexInParentOptimizerList = -1;
    public int IndexInParentOptimizerList {
      get { return indexInParentOptimizerList; }
      set { this.indexInParentOptimizerList = value; }
    }

    public OptimizerJob() : base() {
      
    }
    public OptimizerJob(IOptimizer optimizer)
      : this() {
      this.Optimizer = optimizer;

      if (optimizer is Optimization.Experiment) {
        this.ComputeInParallel = true;
      } else if (optimizer is Optimization.BatchRun) {
        this.ComputeInParallel = false;
      } else {
        this.ComputeInParallel = false;
      }
    }
    [StorableConstructor]
    protected OptimizerJob(bool deserializing) { }
    protected OptimizerJob(OptimizerJob original, Cloner cloner)
      : base(original, cloner) {
      this.Optimizer = cloner.Clone(original.Optimizer);
      this.IndexInParentOptimizerList = original.IndexInParentOptimizerList;
      this.CollectChildJobs = original.CollectChildJobs;
      this.RegisterOptimizerEvents();
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new OptimizerJob(this, cloner);
    }

    [StorableHook(HookType.AfterDeserialization)]
    protected virtual void AfterDeserialization() {
      RegisterOptimizerEvents();
    }

    /// <summary>
    /// Casts the Optimizer to an Experiment. Returns null if cast was not successfull.
    /// </summary>
    public Optimization.Experiment OptimizerAsExperiment {
      get { return Optimizer as Optimization.Experiment; }
    }

    /// <summary>
    /// Casts the Optimizer to an BatchRun. Returns null if cast was not successfull.
    /// </summary>
    public Optimization.BatchRun OptimizerAsBatchRun {
      get { return Optimizer as Optimization.BatchRun; }
    }

    #region IJob Members

    public override ExecutionState ExecutionState {
      get { return optimizer.ExecutionState; }
    }

    public override TimeSpan ExecutionTime {
      get { return optimizer.ExecutionTime; }
    }

    public override void Prepare() {
      optimizer.Prepare();
    }

    public override void Start() {
      if ((optimizer is Optimization.Experiment && OptimizerAsExperiment.Optimizers.Count == 0) || // experiment would not fire OnStopped if it has 0 optimizers
          (optimizer is Optimization.BatchRun && OptimizerAsBatchRun.Optimizer == null)) { // batchrun would not fire OnStopped if algorithm == null
        OnJobStopped();
      } else {
        optimizer.Start();
      }
    }

    public override void Pause() {
      optimizer.Pause();
    }

    public override void Stop() {
      optimizer.Stop();
    }

    public override void Resume(IEnumerable<IJob> childJobs) {
      OnJobStopped();
    }
    #endregion

    #region Optimizer Events
    protected virtual void RegisterOptimizerEvents() {
      optimizer.Stopped += new EventHandler(optimizer_Stopped);
      optimizer.Paused += new EventHandler(optimizer_Paused);
      optimizer.ExceptionOccurred += new EventHandler<EventArgs<Exception>>(optimizer_ExceptionOccurred);
      optimizer.DescriptionChanged += new EventHandler(optimizer_DescriptionChanged);
      optimizer.ItemImageChanged += new EventHandler(optimizer_ItemImageChanged);
      optimizer.NameChanged += new EventHandler(optimizer_NameChanged);
      optimizer.NameChanging += new EventHandler<CancelEventArgs<string>>(optimizer_NameChanging);
      optimizer.ToStringChanged += new EventHandler(optimizer_ToStringChanged);
    }

    protected virtual void DeregisterOptimizerEvents() {
      optimizer.Stopped -= new EventHandler(optimizer_Stopped);
      optimizer.Paused -= new EventHandler(optimizer_Paused);
      optimizer.ExceptionOccurred -= new EventHandler<EventArgs<Exception>>(optimizer_ExceptionOccurred);
      optimizer.DescriptionChanged -= new EventHandler(optimizer_DescriptionChanged);
      optimizer.ItemImageChanged -= new EventHandler(optimizer_ItemImageChanged);
      optimizer.NameChanged -= new EventHandler(optimizer_NameChanged);
      optimizer.NameChanging -= new EventHandler<CancelEventArgs<string>>(optimizer_NameChanging);
      optimizer.ToStringChanged -= new EventHandler(optimizer_ToStringChanged);
    }

    protected void optimizer_ToStringChanged(object sender, EventArgs e) {
      this.OnToStringChanged();
    }

    protected void optimizer_NameChanging(object sender, CancelEventArgs<string> e) {
      this.OnNameChanging(e.Value, e.Cancel);
    }

    protected void optimizer_NameChanged(object sender, EventArgs e) {
      this.OnNameChanged();
    }

    protected void optimizer_ItemImageChanged(object sender, EventArgs e) {
      this.OnItemImageChanged();
    }

    protected void optimizer_DescriptionChanged(object sender, EventArgs e) {
      this.OnDescriptionChanged();
    }

    protected virtual void optimizer_ExceptionOccurred(object sender, EventArgs<Exception> e) {
      OnJobFailed(e);
    }

    protected virtual void optimizer_Stopped(object sender, EventArgs e) {
      OnJobStopped();
    }

    protected virtual void optimizer_Paused(object sender, EventArgs e) {
      OnJobPaused();
    }
    #endregion

    #region INamedItem Members

    public override bool CanChangeDescription {
      get { return optimizer.CanChangeDescription; }
    }

    public override bool CanChangeName {
      get { return optimizer.CanChangeName; }
    }

    public override string Description {
      get { return optimizer.Description; }
      set { optimizer.Description = value; }
    }

    public override string Name {
      get { return optimizer.Name; }
      set { optimizer.Name = value; }
    }
    #endregion
    
    #region IItem Members

    public override string ItemDescription {
      get { return optimizer.ItemDescription; }
    }

    public override Image ItemImage {
      get { return optimizer.ItemImage; }
    }

    public override string ItemName {
      get { return optimizer.ItemName; }
    }

    public override Version ItemVersion {
      get { return optimizer.ItemVersion; }
    }
    #endregion
  }
}
